package extractfreq

import (
	"fmt"
	"log"
	"math"
	"math/cmplx"

	"git.crumpington.com/public/jlaudio/lib/flac"
	"git.crumpington.com/public/jlaudio/lib/util"
	"gonum.org/v1/gonum/fourier"
	"gonum.org/v1/plot"
	"gonum.org/v1/plot/plotter"
	"gonum.org/v1/plot/plotutil"
	"gonum.org/v1/plot/vg"
)

// Steps:
//
// * Create amplitude spectrum (abs of real FFT).
// * Multiply by A-weight curve
// * Find peaks
// * Sort peaks by amplitude
// * Use strongest peak to estimate fFundamental?
// *
func PlotFFT(path string, nSamples int, fFundamental float64) {
	L, R, err := flac.Load(path)
	if err != nil {
		panic(err)
	}

	if nSamples < len(L) {
		nSamples = len(L)
	}

	data := make([]float64, nSamples)
	for i := range data {
		data[i] = float64(L[i]) + float64(R[i])
	}

	// Make initial fft.
	fft := fourier.NewFFT(nSamples)
	cCoeffs := fft.Coefficients(nil, data)
	amps := make([]float64, len(cCoeffs))
	freqs := make([]float64, len(cCoeffs))
	for i := range freqs {
		freqs[i] = fft.Freq(i) * 48000
		amps[i] = cmplx.Abs(cCoeffs[i])
	}

	// Trim to frequencies of interest.
	freq0 := fFundamental / 8.0
	freq1 := 21 * fFundamental
	if freq1 > 20000 {
		freq1 = 20000
	}

	deltaF := freqs[1] - freqs[0]
	idx0 := int(freq0 / deltaF)
	idx1 := int(freq1 / deltaF)
	freqs = freqs[idx0:idx1]
	amps = amps[idx0:idx1]

	// Multiply by A-weight function, then normalize.
	ampMax := float64(0)
	for i := range amps {
		amps[i] *= AWeight(freqs[i])
		if amps[i] > ampMax {
			ampMax = amps[i]
		}
	}

	for i := range amps {
		if amps[i] > 0 {
			amps[i] /= ampMax
		}
	}

	// Find peaks.
	halfWinSize := int(math.Round((fFundamental / deltaF) / 3))
	peakInds, peakAmps := getPeaks(halfWinSize, amps)
	peakFreqs := make([]float64, len(peakInds))
	for i, idx := range peakInds {
		peakFreqs[i] = freqs[idx]
	}

	// First, determine harmonic numbers of each peak.
	ns := make([]float64, len(peakFreqs))
	for i, f := range peakFreqs {
		ns[i] = ApproxHarmonicNumber(fFundamental, f)
	}

	// --------------------------------------------------------------------------

	// First fit.

	log.Printf("First fit...")
	A, B, err := fitSpectrumPeaks(fFundamental, ns, peakFreqs, peakAmps)
	if err != nil {
		panic(err)
	}

	log.Printf("A: %.4f B: %0.4f", A, B)

	p, err := plot.New()
	if err != nil {
		panic(err)
	}

	ptsModel := make(plotter.XYs, len(ns))
	ptsData := make(plotter.XYs, len(ns))
	for i, n := range ns {
		ptsModel[i].X = n
		ptsModel[i].Y = harmonicModel(A, B, n)
		ptsData[i].X = n
		ptsData[i].Y = peakFreqs[i]
	}

	if err = plotutil.AddScatters(p, "model", ptsModel, "data", ptsData); err != nil {
		panic(err)
	}

	p.Add(plotter.NewGrid())

	log.Printf("Saving...")
	if err := p.Save(16*vg.Inch, 8*vg.Inch, "fit1.png"); err != nil {
		panic(err)
	}

	// Remove peaks w/ larger errors:
	// TODO: Method?
	iOut := 0
	for i := range ns {
		f := peakFreqs[i]
		fModel := harmonicModel(A, B, ns[i])
		deltaF := math.Abs(f - fModel)
		maxDeltaF := 20 * util.CentDeltaFreq(fModel)
		if deltaF < maxDeltaF {
			ns[iOut] = ns[i]
			peakInds[iOut] = peakInds[i]
			peakFreqs[iOut] = peakFreqs[i]
			peakAmps[iOut] = peakAmps[i]
			iOut++
		}
	}

	ns = ns[:iOut]
	peakInds = peakInds[:iOut]
	peakFreqs = peakFreqs[:iOut]
	peakAmps = peakAmps[:iOut]

	// Second fit.

	A, B, err = fitSpectrumPeaks(A, ns, peakFreqs, peakAmps)
	if err != nil {
		panic(err)
	}

	log.Printf("A: %.4f B: %0.4f", A, B)

	p, err = plot.New()
	if err != nil {
		panic(err)
	}

	ptsModel = make(plotter.XYs, len(ns))
	ptsData = make(plotter.XYs, len(ns))
	for i, n := range ns {
		ptsModel[i].X = n
		ptsModel[i].Y = harmonicModel(A, B, n)
		ptsData[i].X = n
		ptsData[i].Y = peakFreqs[i]
	}

	if err = plotutil.AddScatters(p, "model", ptsModel, "data", ptsData); err != nil {
		panic(err)
	}

	p.Add(plotter.NewGrid())

	log.Printf("Saving...")
	if err := p.Save(16*vg.Inch, 8*vg.Inch, "fit2.png"); err != nil {
		panic(err)
	}

	// --------------------------------------------------------------------------

	p, err = plot.New()
	if err != nil {
		panic(err)
	}

	log.Printf("Making points...")
	pts := make(plotter.XYs, 0, len(amps))
	for i := range amps {
		pts = append(pts, plotter.XY{
			X: freqs[i],
			Y: math.Log(amps[i]),
		})
	}

	pts2 := make(plotter.XYs, 0, len(amps))
	for i := range peakInds {
		pts2 = append(pts2, plotter.XY{
			X: freqs[peakInds[i]],
			Y: math.Log(peakAmps[i]),
		})
	}

	log.Printf("Plotting...")
	if err = plotutil.AddLines(p, "fft", pts); err != nil {
		panic(err)
	}
	if err = plotutil.AddScatters(p, "max", pts2); err != nil {
		panic(err)
	}
	p.Add(plotter.NewGrid())

	log.Printf("Saving...")
	if err := p.Save(16*vg.Inch, 8*vg.Inch, "fft.png"); err != nil {
		panic(err)
	}

	for i, idx := range peakInds {
		n := ns[i]

		// Make window.
		ampsWin := amps[idx-halfWinSize : idx+halfWinSize]
		freqWin := freqs[idx-halfWinSize : idx+halfWinSize]

		log.Printf("Making points...")
		pts := make(plotter.XYs, 0, 2*halfWinSize)
		for i := range ampsWin {
			pts = append(pts, plotter.XY{
				X: freqWin[i],
				Y: ampsWin[i],
			})
		}

		pts2 := plotter.XYs{
			{
				X: harmonicModel(A, B, n),
				Y: peakAmps[i],
			},
		}

		log.Printf("%f, %f", peakFreqs[i], harmonicModel(A, B, n))

		log.Printf("Plotting...")
		p, err := plot.New()
		if err != nil {
			panic(err)
		}

		if err = plotutil.AddLines(p, "fft", pts); err != nil {
			panic(err)
		}
		if err = plotutil.AddScatters(p, "model", pts2); err != nil {
			panic(err)
		}

		log.Printf("Saving...")
		filename := fmt.Sprintf("plot-%03f.png", n)
		if err := p.Save(8*vg.Inch, 8*vg.Inch, filename); err != nil {
			panic(err)
		}
	}
}
